#include "server.h"
#include "http.h"
#include "send.h"
#include "list.h"

#define CLIENT_ADDR(hc) inet_ntoa((hc)->addr.sin_addr)
#define CLIENT_PORT(hc) (hc)->addr.sin_port
#define HTTP_ERR_MSG "HTTP/1.0 400 Bad Request\n"
extern struct gengetopt_args_info conf;

struct http_client {
	int fd;
	struct sockaddr_in addr;
	int status;
	int check_r, check_w;
	struct list_head node;
};

struct access_info {
	struct in_addr addr;
	int netmask;
	struct list_head node;
};

enum {HTTP_CONNECTED, HTTP_GOT_GET_REQUEST, HTTP_SENT_OK_MSG, HTTP_READY_TO_STREAM,
	HTTP_STREAMING, HTTP_INVALID_GET_REQUEST};

static int server_fd = -1, numclients;
static struct list_head clients, access_perm_list;
static struct sender *self;


static void http_shutdown_client(struct http_client *hc, const char *msg)
{
	PARA_INFO_LOG("shutting down %s on fd %d (%s)\n", CLIENT_ADDR(hc), hc->fd, msg);
	numclients--;
	close(hc->fd);
	list_del(&hc->node);
	free(hc);
	return;
}

static void http_shutdown_clients_real(void)
{
	struct http_client *hc, *tmp;
	list_for_each_entry_safe(hc, tmp, &clients, node)
		http_shutdown_client(hc, "afs request");
}
static void http_shutdown_clients(void)
{
	struct http_client *hc, *tmp;
	list_for_each_entry_safe(hc, tmp, &clients, node)
		if (hc->status == HTTP_STREAMING)
			http_shutdown_client(hc, "afs request");
}

static int http_send_msg(struct http_client *hc, const char *msg)
{
	int ret = send_va_buffer(hc->fd, "%s", msg);

	if (ret < 0)
		http_shutdown_client(hc, "send msg failed");
	else
		PARA_INFO_LOG("sent message: %s\n", msg);
	return ret;
}

static void http_send_ok_msg(struct http_client *hc)
{
	http_send_msg(hc, HTTP_OK_MSG);
}

static int http_send_err_msg(struct http_client *hc)
{
	return http_send_msg(hc, HTTP_ERR_MSG);
}

static int http_send(struct audio_format *af, long unsigned current_chunk,
		long unsigned chunks_sent, const char *buf, size_t len)
{
	fd_set wfds;
	struct http_client *hc, *tmp;
	int i, ret, max = -1;
	struct timeval tv = {0, 0};

	if (server_fd <= 0)
		return -1;
	FD_ZERO(&wfds);
	list_for_each_entry(hc, &clients, node) {
		FD_SET(hc->fd, &wfds);
		max = MAX(max, hc->fd);
	}
	if (max < 0)
		return 1;
	ret = select(max + 1, NULL, &wfds, NULL, &tv);
	if (ret < 0)
		return -1;
	i = -1;
	list_for_each_entry_safe(hc, tmp, &clients, node) {
		i++;
		if (hc->status != HTTP_STREAMING)
			continue;
//		PARA_DEBUG_LOG("sending to client #%d %s, status = %d\n", i, CLIENT_ADDR(hc), hc->status);
		if (!FD_ISSET(hc->fd, &wfds)) {
			http_shutdown_client(hc, "fd not ready for writing");
			continue;
		}
//		PARA_DEBUG_LOG("sending %d bytes to client %s\n", len, CLIENT_ADDR(hc));
		if (send_bin_buffer(hc->fd, buf, len) < 0) {
			http_shutdown_client(hc, "send buffer failed");
			continue;
		}
	}
	return 1;
}

/* FIXME: There must be a better way */
static int http_check_perms(struct http_client *hc)
{
	struct access_info *ai, *tmp;
	char *nad_hc = inet_ntoa(hc->addr.sin_addr);
	unsigned x, x1, x2, x3, x4;

	if (sscanf(nad_hc, "%d.%d.%d.%d", &x1, &x2, &x3, &x4) != 4)
		return 0; /* weird */
	x = (x1 << 24) + (x2 << 16) + (x3 << 8) + x4;

	list_for_each_entry_safe(ai, tmp, &access_perm_list, node) {
		char *nad = inet_ntoa(ai->addr);
		unsigned y, y1, y2, y3, y4, mask = ((unsigned)(~0) <<
			(32 - ai->netmask));
//		PARA_DEBUG_LOG("checking if %s matches %s/%d\n", nad_hc, nad, ai->netmask);
		if (sscanf(nad, "%d.%d.%d.%d", &y1, &y2, &y3, &y4) != 4)
			continue;
		y = (y1 << 24) + (y2 << 16) + (y3 << 8) + y4;
//		PARA_DEBUG_LOG("x = %u, y=%u\n", x, y);
		if ((x & (1 << mask)) == (y & (1 << mask)))
			return 1; /* permission denied */
	}
	return 0;
}

static int http_send_header(struct audio_format *af, struct http_client *hc)
{
	int len;
	char *buf = af->get_header_info(&len);
	if (!buf || len <= 0)
		return -1;
	if (send_bin_buffer(hc->fd, buf, len) < 0) {
		http_shutdown_client(hc, "failed to send header");
		return -1;
	}
	return 1;
}

static void http_handle_fds(struct audio_format *af, int *max_fileno,
		fd_set *rfds, fd_set *wfds)
{
	int i = -1, match;
	struct http_client *hc, *tmp;
	char *err_msg;

	list_for_each_entry_safe(hc, tmp, &clients, node) {
		i++;
//		PARA_DEBUG_LOG("handling client %d: %s\n", i, CLIENT_ADDR(hc));
		switch (hc->status) {
		case HTTP_STREAMING: /* nothing to do */
			break;
		case HTTP_READY_TO_STREAM:
			if (hc->check_w && FD_ISSET(hc->fd, wfds)) {
				PARA_INFO_LOG("client #%d on fd %d (%s): "
					"streaming\n", i, hc->fd,
					CLIENT_ADDR(hc));
				hc->status = HTTP_STREAMING;
			}
			break;
		case HTTP_CONNECTED: /* need to recv get request */
			if (hc->check_r && FD_ISSET(hc->fd, rfds)) {
				if (recv_pattern(hc->fd, HTTP_GET_MSG, MAXLINE) < 0) {
					hc->status = HTTP_INVALID_GET_REQUEST;
				} else {
					hc->status = HTTP_GOT_GET_REQUEST;
					PARA_INFO_LOG("%s", "received get request\n");
				}
			}
			break;
		case HTTP_GOT_GET_REQUEST: /* need to send ok msg */
			if (hc->check_w && FD_ISSET(hc->fd, wfds)) {
				hc->status = HTTP_SENT_OK_MSG;
				http_send_ok_msg(hc);
			}
			break;
		case HTTP_INVALID_GET_REQUEST: /* need to send err msg */
			if (hc->check_w && FD_ISSET(hc->fd, wfds)) {
				if (http_send_err_msg(hc) >= 0);
					http_shutdown_client(hc, "invalid get request");
			}
			break;
		case HTTP_SENT_OK_MSG: /* need to send header? */
			if (hc->check_w && FD_ISSET(hc->fd, wfds)) {
				if (!af)
					break;
				hc->status = HTTP_READY_TO_STREAM;
				if (!af->get_header_info)
					break; /* this af needs no headers */
				if (http_send_header(af, hc) > 0)
					break;;
				hc->status = HTTP_SENT_OK_MSG; /* try again */
			}
			break;
		}
	}
	if (!FD_ISSET(server_fd, rfds))
		return;
	hc = para_malloc(sizeof(struct http_client));
	err_msg = "accept error";
	hc->fd = do_accept(server_fd, &hc->addr);
	if (hc->fd <= 0)
		goto err_out;
	PARA_INFO_LOG("connection from %s (fd %d)\n", CLIENT_ADDR(hc), hc->fd);
	if (conf.http_max_clients_arg > 0 && numclients >=
			conf.http_max_clients_arg) {
		err_msg = "server full";
		goto err_out;
	}
	match = http_check_perms(hc);
	if ((match && !conf.http_default_deny_given) ||
			(!match && conf.http_default_deny_given)) {
		err_msg = "permission denied";
		goto err_out;
	}
	hc->status = HTTP_CONNECTED;
	hc->check_r = 0;
	hc->check_w = 0;
	PARA_INFO_LOG("accepted client #%d: %s (fd %d)\n", numclients, CLIENT_ADDR(hc), hc->fd);
	numclients++;
	list_add(&hc->node, &clients);
	return;
err_out:
	PARA_WARNING_LOG("ignoring connect request from %s (%s)\n", CLIENT_ADDR(hc),
		err_msg);
	if (hc->fd > 0)
		close(hc->fd);
	free(hc);
}

static int http_set_fds(struct audio_format *af, int *max_fileno, fd_set *rfds, fd_set *wfds)
{
	struct http_client *hc, *tmp;
	int i = -1;;

//	PARA_INFO_LOG("FD_SET: %d\n", server_fd);
	FD_SET(server_fd, rfds);
	*max_fileno = MAX(*max_fileno, server_fd);
	list_for_each_entry_safe(hc, tmp, &clients, node) {
		i++;
//		PARA_DEBUG_LOG("checking fds for client %d: %s\n", i, CLIENT_ADDR(hc));
		hc->check_r = 0;
		hc->check_w = 0;
		switch (hc->status) {
		case HTTP_STREAMING:
			break;
		case HTTP_CONNECTED: /* need to recv get request */
			FD_SET(hc->fd, rfds);
			*max_fileno = MAX(*max_fileno, hc->fd);
			hc->check_r = 1;
			break;
		case HTTP_READY_TO_STREAM: /* need to wait until fd is writable again */
		case HTTP_GOT_GET_REQUEST: /* need to send ok msg */
		case HTTP_INVALID_GET_REQUEST: /* need to send err msg */
			FD_SET(hc->fd, wfds);
			*max_fileno = MAX(*max_fileno, hc->fd);
			hc->check_w = 1;
			break;
		case HTTP_SENT_OK_MSG:
			if (!af || !afs_playing())
				break; /* wait until server starts playing */
			if (af->get_header_info) { /* need to send header */
				FD_SET(hc->fd, wfds);
				*max_fileno = MAX(*max_fileno, hc->fd);
				hc->check_w = 1;
			} else /* this af needs no header send */
				hc->status = HTTP_READY_TO_STREAM;
			break;
		}
	}
	return 1;
}

static int open_tcp_port(int port)
{
	int flags;

	server_fd = init_tcp_socket(port);
	if (server_fd < 0) {
		PARA_CRIT_LOG("failed to init tcp port %d\n", port);
		http_shutdown_clients_real();
		self->status = SENDER_OFF;
		return -1;
	}
	self->status = SENDER_ON;
	flags = fcntl(server_fd, F_GETFL);
	fcntl(server_fd, F_SETFD, FD_CLOEXEC);
	return 1;
}

static int http_com_on(struct sender_command_data *scd)
{
	if (self->status == SENDER_ON)
		return 1;
	return open_tcp_port(conf.http_port_arg);
}

static int http_com_off(struct sender_command_data *scd)
{
	self->status = SENDER_OFF;
	if (server_fd > 0) {
		close(server_fd);
		server_fd = -1;
	}
	http_shutdown_clients_real();
	return 1;
}

static int del_perm_list_entry(struct sender_command_data *scd)
{
	struct access_info *ai, *tmp;

	list_for_each_entry_safe(ai, tmp, &access_perm_list, node) {
		char *nad = para_strdup(inet_ntoa(ai->addr));
		if (!strcmp(nad, inet_ntoa(scd->addr)) &&
				ai->netmask == scd->netmask) {
			PARA_NOTICE_LOG("removing %s/%i from access list\n",
				nad, ai->netmask);
			list_del(&ai->node);
			free(ai);
		}
		free(nad);
	}
	return 1;
}

static int add_perm_list_entry(struct sender_command_data *scd)
{
	struct access_info *ai = para_malloc(sizeof(struct access_info));
	ai->addr = scd->addr;
	ai->netmask = scd->netmask;
	PARA_INFO_LOG("adding %s/%i to access list\n", inet_ntoa(ai->addr),
		ai->netmask);
	list_add(&ai->node, &access_perm_list);
	return 1;
}

static int http_com_deny(struct sender_command_data *scd)
{
	return conf.http_default_deny_given?
		del_perm_list_entry(scd) : add_perm_list_entry(scd);
}

static int http_com_allow(struct sender_command_data *scd)
{
	return conf.http_default_deny_given?
	add_perm_list_entry(scd) : del_perm_list_entry(scd);
}

static char *http_info(void)
{
	char *clnts = NULL, *ap = NULL, *ret;
	struct access_info *ai, *tmp_ai;
	struct http_client *hc, *tmp_hc;

	list_for_each_entry_safe(ai, tmp_ai, &access_perm_list, node) {
		char *tmp = make_message("%s%s/%d ", ap? ap : "",
			inet_ntoa(ai->addr), ai->netmask);
		free(ap);
		ap = tmp;
	}
	list_for_each_entry_safe(hc, tmp_hc, &clients, node) {
		char *tmp = make_message("%s%s:%d ", clnts? clnts : "",
			CLIENT_ADDR(hc), CLIENT_PORT(hc));
		free(clnts);
		clnts = tmp;
	}
	ret = make_message(
		"http status: %s\n"
		"http tcp port: %d\n"
		"http clients: %d\n"
		"http maximal number of clients: %d%s\n"
		"http connected clients: %s\n"
		"http access %s list: %s\n",
		(self->status == SENDER_ON)? "on" : "off",
		conf.http_port_arg,
		numclients,
		conf.http_max_clients_arg,
		conf.http_max_clients_arg > 0? "" : " (unlimited)",
		clnts? clnts : "(none)",
		conf.http_default_deny_given? "allow" : "deny",
		ap? ap : "(none)"
	);
	free(ap);
	free(clnts);
	return ret;
}

static int init_access_control_list(void)
{
	int i;
	struct sender_command_data scd;

	INIT_LIST_HEAD(&access_perm_list);
	for (i = 0; i < conf.http_access_given; i++) {
		char *arg = para_strdup(conf.http_access_arg[i]);
		char *p = strchr(arg, '/');
		if (!p)
			goto err;
		*p = '\0';
		if (!inet_aton(arg, &scd.addr))
			goto err;
		scd.netmask = atoi(++p);
		if (scd.netmask < 0 || scd.netmask > 32)
			goto err;
		add_perm_list_entry(&scd);
		free(arg);
		continue;
err:
		PARA_CRIT_LOG("syntax error for http_access option #%d\n", i);
		free(arg);
		return -1;
	}
	return 1;
}

static char *http_help(void)
{
	return make_message(
		"usage: {on|off}\n"
		"usage: {allow|deny} IP mask\n"
		"example: allow 127.0.0.1 32\n"
	);
}

int para_http_init(struct sender *s)
{
	INIT_LIST_HEAD(&clients);
	s->info = http_info;
	s->send = http_send;
	s->set_fds = http_set_fds;
	s->handle_fds = http_handle_fds;
	s->shutdown_clients = http_shutdown_clients;
	s->help = http_help;
	s->client_cmds[SENDER_ON] = http_com_on;
	s->client_cmds[SENDER_OFF] = http_com_off;
	s->client_cmds[SENDER_DENY] = http_com_deny;
	s->client_cmds[SENDER_ALLOW] = http_com_allow;
	s->client_cmds[SENDER_ADD] = NULL;
	s->client_cmds[SENDER_DELETE] = NULL;
	self = s;
	if (init_access_control_list() < 0)
		return -1;
	if (!conf.http_no_autostart_given)
		if (open_tcp_port(conf.http_port_arg) < 0)
			return -1;
	PARA_DEBUG_LOG("%s", "http sender init complete\n");
	return 1;
}
